'use strict';

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var _ = require('lodash');
var React = require('react');
var PropTypes = require('./prop-types');

var options = void 0;
var initialized = false;

var addHiddenProperties = function addHiddenProperties(obj, prop, value) {
  addHiddenProperty(obj, prop, value);
  addHiddenProperty(obj.isRequired, prop, value);
  return obj;
};

var addHiddenProperty = function addHiddenProperty(obj, prop, value) {
  return Object.defineProperty(obj, prop, { enumerable: false, value: value });
};

var wrapPropTypes = function wrapPropTypes() {
  initialized = true;
  // Adds a .type key which allows the type to be derived during the
  // evaluation process. This is necessary for complex types which
  // return the result of a generator function, leaving no way to
  // determine which type instantiated it.

  var original = _.cloneDeep(PropTypes);

  _.each(_.keys(GENERATORS), function (k) {
    if (PropTypes[k] === undefined) return;
    if (PropTypes[k].isRequired !== undefined) {
      // Simple type. Just extend the object
      PropTypes[k] = addHiddenProperties(PropTypes[k], 'type', k);
    } else {
      // Complex type. Must extend the creator's return value
      PropTypes[k] = function (arg) {
        var res = original[k](arg);
        res = addHiddenProperties(res, 'type', k);
        res = addHiddenProperties(res, 'arg', arg);
        return res;
      };
    }
  });
};

var GENERATORS = {
  // Simple types
  array: function array(propName) {
    return propName ? [propName] : [];
  },
  bool: function bool() {
    return true;
  },
  func: function func() {
    return function () {};
  },
  number: function number() {
    return 1;
  },
  object: function object(propName) {
    return propName ? _defineProperty({}, propName, propName) : {};
  },
  string: function string(propName) {
    return propName || 'string';
  },
  any: function any(propName) {
    return propName || 'any';
  },
  element: function element(propName) {
    return React.createElement('div', propName);
  },
  node: function node(propName) {
    return propName || 'node';
  },
  symbol: function symbol(propName) {
    return Symbol(propName);
  },

  // Complex types
  arrayOf: function arrayOf(propName, type) {
    var res = generateOneProp(type, propName, false);
    return res ? [res] : [];
  },
  instanceOf: function instanceOf(propName, klass) {
    return new klass();
  },
  objectOf: function objectOf(propName, type) {
    return { key: generateOneProp(type, propName, false) };
  },
  oneOf: function oneOf(propName, values) {
    return _.first(values);
  },
  oneOfType: function oneOfType(propName, types) {
    return forceGenerateOneProp(_.first(types), propName);
  },
  shape: function shape(propName, _shape) {
    return generateProps(_shape, options);
  }
};

var shouldGenerate = function shouldGenerate(propType) {
  return (
    // Generate required props, and this is the required version
    options.required && !propType.isRequired ||
    // Generate optional props, and this is the optional version
    options.optional && !!propType.isRequired
  );
};

var initError = new Error('generateProps.init() must be called at the beginning of your test suite');

var generateOneProp = function generateOneProp(propType, propName) {
  var wrapInArray = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;

  if (propType.type === undefined && initialized === false) {
    throw initError;
  }

  var generate = options.generators[propType.type].bind(undefined, propName);
  var arg = propType.arg;
  if (generate) {
    if (shouldGenerate(propType)) {
      if (wrapInArray) {
        return [propName, generate(arg)];
      } else {
        return generate(arg);
      }
    }
  }
};

var forceGenerateOneProp = function forceGenerateOneProp(propType, propName) {
  var generate = GENERATORS[propType.type].bind(undefined, propName);
  var arg = propType.arg;
  if (generate) {
    return generate(arg);
  }
};

var generateProps = function generateProps(arg, opts) {
  if (initialized === false) {
    throw initError;
  }

  options = _.defaults({}, opts, { required: true, optional: false });
  if (opts && opts.generators) {
    options.generators = _.defaults({}, opts.generators, GENERATORS);
  } else {
    options.generators = GENERATORS;
  }

  var propTypes = void 0;

  if (!arg) {
    throw new TypeError('generateProps expected a propType object or a React Component');
  }

  if (arg.type && GENERATORS[arg.type]) {
    return generateOneProp(arg, undefined, false);
  } else if (arg.propTypes && arg.propTypes.type && GENERATORS[arg.propTypes.type]) {
    return generateOneProp(arg.propTypes, undefined, false);
  }

  if (_.isPlainObject(arg.propTypes)) {
    propTypes = arg.propTypes;
  } else if (_.isPlainObject(arg)) {
    propTypes = arg;
  } else {
    throw new TypeError('generateProps expected a propType object or a React Component');
  }

  return _(propTypes).map(generateOneProp).compact().fromPairs().value();
};

Object.assign(generateProps, { init: wrapPropTypes });

module.exports = generateProps;